@@ -11,7 +11,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
 from datetime import datetime, timezone
 from os.path import basename, dirname
 from pathlib import Path
@@ -43,6 +42,15 @@ def test_ls(fs: V3ioFS, new_file, path_cls):
     assert all(isinstance(p, str) for p in out), 'not string'
 
 
+def test_ls_with_marker(fs: V3ioFS, new_file):
+    for i in range(1200):
+        new_file(fs._client, f'{test_dir}/test_ls/test-file{i}')
+    path = str(f'/{test_container}/{test_dir}/test_ls')
+
+    out = fs.ls(path, detail=True)
+    assert len(out) == 1200, 'not all files returned'
+
+
 def test_rm(fs: V3ioFS, tmp_obj):
     path = tmp_obj.path
     fs.rm(path)
@@ -116,5 +124,4 @@ def test_fsspec():
     with fs.open(prefix + filepath) as fp:
         data = fp.read()
     assert data == b'123', 'unexpected data'
-    print(filepath)
     fs.rm(filepath)
