@@ -11,10 +11,9 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
-
 from datetime import datetime, timezone
 from os import environ
+import time
 from urllib.parse import urlparse
 
 from fsspec.spec import AbstractFileSystem
@@ -29,6 +28,57 @@ _file_key = 'key'
 _dir_key = 'prefix'
 
 
+class _Cache:
+    def __init__(self, capacity, cache_validity_seconds):
+        self._cache = {}
+        self._time_to_key = []
+        self._capacity = capacity
+        self._cache_validity_seconds = cache_validity_seconds
+
+    def put(self, key, value):
+        start_time = time.monotonic()
+
+        self._cache[key] = (start_time, value)
+        self._time_to_key.append((start_time, key))
+
+        # GC
+        if len(self._cache) > self._capacity:
+            self._gc(start_time)
+
+    def get(self, key):
+        start_time = time.monotonic()
+
+        lookup_result = self._cache.get(key)
+        if lookup_result is None:
+            return None
+        key_time, value = lookup_result
+        if key_time <= start_time + self._cache_validity_seconds:
+            return value
+
+        self._gc(start_time)
+
+        return None
+
+    def delete_if_exists(self, key):
+        self._cache.pop(key, None)
+
+    def _gc(self, until):
+        for key_time, key in self._time_to_key:
+            num_removed = 0
+            if key_time > until + self._cache_validity_seconds or len(self._cache) > self._capacity:
+                result = self._cache.get(key, None)
+                if result:
+                    cache_time, value = result
+                    # cache entry may have been deleted and re-added
+                    if cache_time == key_time:
+                        del self._cache[key]
+
+                num_removed += 1
+            else:
+                break
+            self._time_to_key = self._time_to_key[num_removed:]
+
+
 class V3ioFS(AbstractFileSystem):
     """File system driver to v3io
 
@@ -38,67 +88,94 @@ class V3ioFS(AbstractFileSystem):
         API host name (or V3IO_API environment)
     v3io_access_key: str
         v3io access key (or V3IO_ACCESS_KEY from environment)
+    cache_validity_seconds: int | str | None
+        use caching for info(), with invalidation after cache_validity_seconds. Default is 2. Set to 0 to disable.
+    cache_capacity: int | str | None
+        limits the size of the cache. If cache_validity_seconds is not set, this parameter has no effect. Default is 128.
     **kw:
         Passed to fsspec.AbstractFileSystem
     """
 
     protocol = 'v3io'
 
-    def __init__(self, v3io_api=None, v3io_access_key=None, **kw):
+    def __init__(self, v3io_api=None, v3io_access_key=None, cache_validity_seconds=None, cache_capacity=None, **kw):
         # TODO: Support storage options for creds (in kw)
         super().__init__(**kw)
         self._client = _new_client(v3io_api, v3io_access_key)
+        self._cache = None
+        if cache_validity_seconds is None:
+            cache_validity_seconds = 2
+        if cache_capacity is None:
+            cache_capacity = 128
+        if cache_validity_seconds > 0:
+            self._cache = _Cache(int(cache_capacity), int(cache_validity_seconds))
 
-    def ls(self, path, detail=True, **kwargs):
+    def ls(self, path, detail=True, marker=None, **kwargs):
         """Lists files & directories under path"""
+
         full_path = path
         path = strip_schema(path)
         container, path = split_container(path)
         if not container:
             return self._list_containers(detail)
+        ext_out = []
 
-        resp = self._client.get_container_contents(
-            container=container,
-            path=path,
-            get_all_attributes=True,
-            raise_for_status=v3io.dataplane.RaiseForStatus.never,
-        )
+        while True:
+            resp = self._client.get_container_contents(
+                container=container,
+                path=path,
+                get_all_attributes=True,
+                raise_for_status=v3io.dataplane.RaiseForStatus.never,
+                marker=marker,
+            )
 
-        # Ignore 404's here
-        if resp.status_code not in {200, 404}:
-            raise Exception(
-                f'{resp.status_code} received while accessing {path!r}')
+            # Ignore 404's here
+            if resp.status_code not in {200, 404}:
+                raise Exception(
+                    f'{resp.status_code} received while accessing {path!r}')
 
-        # If not data, try to find file in parent directory
-        if not _has_data(resp):
-            return [self._ls_file(container, path, detail)]
+            out = (
+                    _resp_dirs(resp, container, detail) +
+                    _resp_files(resp, container, detail)
+            )
 
-        out = (
-                _resp_dirs(resp, container, detail) +
-                _resp_files(resp, container, detail)
-        )
+            if not marker:
+                # first time in
+                if not _has_data(resp):
+                    return [self._ls_file(container, path, detail)]
+                if not out:
+                    raise FileNotFoundError(f'{full_path!r} not found')
 
-        if not out:
-            raise FileNotFoundError(f'{full_path!r} not found')
+            ext_out.extend(out)
+            if hasattr(resp.output, 'next_marker') and resp.output.next_marker:
+                marker = resp.output.next_marker
+            else:
+                break
+        return ext_out
 
-        return out
-
-    def _ls_file(self, container, path, detail):
+    def _ls_file(self, container, path, detail, marker=None):
         # '/a/b/c' -> ('/a/b', 'c')
         dirname, _, filename = path.rpartition('/')
-        resp = self._client.get_container_contents(
-            container=container,
-            path=dirname,
-            get_all_attributes=True,
-            raise_for_status=v3io.dataplane.RaiseForStatus.never,
-        )
+        while True:
+            resp = self._client.get_container_contents(
+                container=container,
+                path=dirname,
+                get_all_attributes=True,
+                raise_for_status=v3io.dataplane.RaiseForStatus.never,
+                marker=marker,
+            )
 
-        full_path = f'/{container}/{path}'
-        handle_v3io_errors(resp, full_path)
-        contents = getattr(resp.output, 'contents', [])
-        objs = [obj for obj in contents if path_equal(obj.key, path)]
-        if not objs:
-            raise FileNotFoundError(full_path)
+            full_path = f'/{container}/{path}'
+            handle_v3io_errors(resp, full_path)
+            contents = getattr(resp.output, 'contents', [])
+            objs = [obj for obj in contents if path_equal(obj.key, path)]
+            if not objs:
+                if hasattr(resp.output, 'next_marker') and resp.output.next_marker:
+                    marker = resp.output.next_marker
+                else:
+                    raise FileNotFoundError(full_path)
+            else:
+                break
 
         obj = objs[0]
         if not detail:
@@ -116,13 +193,13 @@ class V3ioFS(AbstractFileSystem):
         ...  # FIXME
 
     def _rm(self, path):
-        container, path = split_container(path)
+        container, path_without_container = split_container(path)
         if not container:
             raise ValueError(f'bad path: {path:r}')
 
         resp = self._client.delete_object(
             container=container,
-            path=path,
+            path=path_without_container,
             raise_for_status=v3io.dataplane.RaiseForStatus.never,
         )
 
@@ -131,6 +208,9 @@ class V3ioFS(AbstractFileSystem):
             raise Exception(
                 f'{resp.status_code} received while accessing {path!r}')
 
+        if self._cache:
+            self._cache.delete_if_exists(path)
+
     def touch(self, path, truncate=True, **kwargs):
         if not truncate:  # TODO
             raise ValueError('only truncate touch supported')
@@ -164,19 +244,57 @@ class V3ioFS(AbstractFileSystem):
             directory, or something else) and other FS-specific keys.
         """
 
-        path = strip_schema(path)
-        out = self.ls(path, detail=True, **kw)
-        entries = [o for o in out if path_equal(o['name'], path)]
+        path_with_container = strip_schema(path)
 
-        if len(entries) == 1:
-            entry = entries[0]
-            entry.setdefault('size', None)
+        if self._cache:
+            lookup_result = self._cache.get(path_with_container)
+            if lookup_result:
+                return lookup_result
+
+        container, path_without_container = split_container(path_with_container)
+
+        # First, we try to get the file's attributes, which will fail with a 404 if it's actually a directory.
+        resp = self._client.get_item(container,
+                                     path_without_container,
+                                     attribute_names=['__size', '__mtime_secs', '__mtime_nsecs', '__mode', '__gid', '__uid'],
+                                     raise_for_status=v3io.dataplane.RaiseForStatus.never)
+
+        if resp.status_code == 200:
+            mtime = int(resp.output.item['__mtime_secs']) + int(resp.output.item['__mtime_nsecs']) / 10 ** 9
+            entry = {
+                'name': path_with_container,
+                'type': 'file',
+                'size': resp.output.item['__size'],
+                'mtime': mtime,
+                'mode': resp.output.item['__mode'],
+                'gid': resp.output.item['__gid'],
+                'uid': resp.output.item['__uid'],
+            }
+            if self._cache:
+                self._cache.put(path_with_container, entry)
             return entry
+        elif resp.status_code == 404:
+            pass  # The file may still be a directory.
+        else:
+            raise Exception(f'{resp.status_code} received while getting the attributes of {path_with_container!r}')
 
-        if len(entries) >= 1 or out:
-            return {'name': path, 'size': 0, 'type': 'directory'}
+        # Check the existence of a directory at the provided path.
+        resp = self._client.get_container_contents(
+            container=container,
+            path=path_without_container,
+            raise_for_status=v3io.dataplane.RaiseForStatus.never,
+            limit=0
+        )
 
-        raise FileNotFoundError(path)
+        if resp.status_code == 200:
+            entry = {'name': path_with_container, 'size': 0, 'type': 'directory'}
+            if self._cache:
+                self._cache.put(path_with_container, entry)
+            return entry
+        elif resp.status_code == 404:
+            raise FileNotFoundError(path_with_container)
+        else:
+            raise Exception(f'{resp.status_code} received while listing {path_with_container!r}')
 
     def _open(
             self,
